unit HrastUnit;

// Copyright  2014 by HrastProgrammer. All rights reserved.
// This unit contains WREN synthesizer modules developed by HrastProgrammer.
// Can be distributed only as part of the complete WREN source package.

interface

uses KnobsUtils, KnobsConversions, Module_Defs, Module_Implementations;

const TableSize=8192;

type

  THrastOsc=class(TMod)
  protected
    z,Phase: TSignal;
    PrevSync: Boolean;
  public
    procedure SetDefaults; override;
    procedure DoTick; override;
    procedure GatherLights(const Prefix: string; const Callback: TLightsHandler); override;
    function InputName(Index: Integer): string; override;
    function OutputName(Index: Integer): string; override;
  end;

  THrastSVF=class(TMod)
  protected
    Band12,Low12,Band24,Low24: TSignal;
  public
    procedure SetDefaults; override;
    procedure DoTick; override;
    function InputName(Index: Integer): string; override;
    function OutputName(Index: Integer): string; override;
  end;

var Pow2Table,SineTable: array [0..TableSize+1] of TSignal;

function CreateModule(ModuleType: Integer; Parent: TSynthPatch; const Name: string): TMod;
procedure CreateTables; // BLue Hell : remaed inintialize and made it public; table re-init needed after changing the system rate

implementation

uses Globals, Math, SysUtils;

const HrastModules: array [4001..4002] of TModuleClass=(THrastOsc,THrastSVF);

function CreateModule;
begin
  if (ModuleType<Low(HrastModules)) or (ModuleType>High(HrastModules)) then
    Result:=nil
  else
    Result:=HrastModules[ModuleType].Create(Parent,Name)
end;

function Pow2Freq(Freq: TSignal): TSignal;
var
  n: Integer;
  f: TSignal;
begin
  f:=Abs(Freq)*TableSize;
  if f>TableSize-1 then f:=TableSize-1;
  n:=Trunc(f); f:=f-n;
  Result:=Pow2Table[n];
  Result:=Result+f*(Pow2Table[n+1]-Result)
end;

function Sine(Freq: TSignal): TSignal;
var
  n: Integer;
  f: TSignal;
begin
  f:=Abs(Freq)*TableSize;
  if f>TableSize-1 then f:=TableSize-1;
  n:=Trunc(f); f:=f-n;
  Result:=SineTable[n];
  Result:=Result+f*(SineTable[n+1]-Result)
end;

{ THrastOsc }

const
  HrastOsc_Freq=0;
  HrastOsc_Morph=1;
  HrastOsc_Shape=2;
  HrastOsc_Bandlimit=3;
  HrastOsc_PW=4;
  HrastOsc_PWM=5;
  HrastOsc_PWMInput=6;
  HrastOsc_FM=7;
  HrastOsc_FMInput=8;
  HrastOsc_PM=9;
  HrastOsc_PMInput=10;
  HrastOsc_Vintage=11;
  HrastOsc_SyncInput=12;
  HrastOsc_Level=13;
  HrastOsc_Mute=14;
  HrastOsc_Cents=15;

procedure THrastOsc.GatherLights;
begin
  Callback(Self,Format('%s/%s/SyncInput/%f',[Prefix,Name,FInCache[HrastOsc_SyncInput]],AppLocale))
end;

function THrastOsc.InputName;
begin
  case Index of
    HrastOsc_Freq: Result:='Freq';
    HrastOsc_Morph: Result:='Morph';
    HrastOsc_Shape: Result:='Shape';
    HrastOsc_Bandlimit: Result:='Bandlimit';
    HrastOsc_PW: Result:='PW';
    HrastOsc_PWM: Result:='PWM';
    HrastOsc_PWMInput: Result:='PWMInput';
    HrastOsc_FM: Result:='FM';
    HrastOsc_FMInput: Result:='FMInput';
    HrastOsc_PM: Result:='PM';
    HrastOsc_PMInput: Result:='PMInput';
    HrastOsc_Vintage: Result:='Vintage';
    HrastOsc_SyncInput: Result:='SyncInput';
    HrastOsc_Level: Result:='Level';
    HrastOsc_Mute: Result:='Mute';
    HrastOsc_Cents: Result:='Cents'
    else Result:=No_Input
  end
end;

function THrastOsc.OutputName;
begin
  case Index of
    0: Result:='Output';
    else Result:=No_Output
  end
end;

procedure THrastOsc.SetDefaults;
begin
  FInCache[HrastOsc_FMInput]:=0;
  FInCache[HrastOsc_PMInput]:=0;
  FInCache[HrastOsc_PWMInput]:=0;
  FKnobmap:=[
    HrastOsc_Freq,
    HrastOsc_Morph,
    HrastOsc_Shape,
    HrastOsc_Bandlimit,
    HrastOsc_PW,
    HrastOsc_PWM,
    HrastOsc_FM,
    HrastOsc_PM,
    HrastOsc_Level
  ]
end;

procedure THrastOsc.DoTick;
var DeltaPhase: TSignal;

  function PolyBLEP(t: TSignal): TSignal;
  var dt: TSignal;
  begin
    dt:=FInCache[HrastOsc_Bandlimit]*DeltaPhase;
    if t<dt then
    begin
      t:=t/dt-1;
      Result:=-t*t
    end
    else
      if t>1-dt then
      begin
        t:=(t-1)/dt+1;
        Result:=t*t
      end
      else
        Result:=0
  end;

var
  n: Integer;
  t,y,ypw: TSignal;
begin
  if not PrevSync and SignalToLogic(FInCache[HrastOsc_SyncInput]) then Phase:=0;
  PrevSync:=SignalToLogic(FInCache[HrastOsc_SyncInput]);
  if SignalToLogic(FInCache[HrastOsc_Mute]) then begin FOutCache[0]:=0; Exit end;
  DeltaPhase:=Pow2Freq(FInCache[HrastOsc_Freq]+FinCache[HrastOsc_Cents]+FInCache[HrastOsc_FM]*FInCache[HrastOsc_FMInput])/System_Rate;
  t:=Phase+FInCache[HrastOsc_PM]*FInCache[HrastOsc_PMInput];
  n:=Trunc(t); t:=t-n; if t<0 then t:=t+1;
  Phase:=Phase+DeltaPhase;
  n:=Trunc(Phase); Phase:=Phase-n; if Phase<0 then Phase:=Phase+1;
  y:=2*t-1-PolyBlep(t);
  t:=t+FInCache[HrastOsc_PW]+FInCache[HrastOsc_PWM]*FInCache[HrastOsc_PWMInput];
  n:=Trunc(t); t:=t-n; if t<0 then t:=t+1;
  ypw:=2*t-1-PolyBlep(t);
  t:=FInCache[HrastOsc_Shape];
  if FInCache[HrastOsc_Vintage]<>0 then
  begin
    if t>0.999 then t:=0.999;
    y:=(y+1)*0.5; y:=y*y; y:=2*y-1;
    ypw:=(ypw+1)*0.5; ypw:=ypw*ypw; ypw:=2*ypw-1
  end;
  z:=y-ypw*FInCache[HrastOsc_Morph]+t*z;
  if t>0.99 then t:=0.99;
  y:=z*(t-1);
  FOutCache[0]:=y*FInCache[HrastOsc_Level]
end;

{ THrastSVF }

const
  HrastSVF_Input=0;
  HrastSVF_Cutoff=1;
  HrastSVF_Reso=2;
  HrastSVF_CutoffInput=3;
  HrastSVF_CutoffMod=4;
  HrastSVF_ResoInput=5;
  HrastSVF_ResoMod=6;
  HrastSVF_Gain=7;
  HrastSVF_Drive=8;
  HrastSVF_LowPass=9;
  HrastSVF_BandPass=10;
  HrastSVF_HighPass=11;
  HrastSVF_Level12=12;
  HrastSVF_Level24=13;
  HrastSVF_OutputLP=0;
  HrastSVF_OutputBP=1;
  HrastSVF_OutputHP=2;
  HrastSVF_OutputBR=3;
  HrastSVF_Output12=4;
  HrastSVF_Output24=5;

function THrastSVF.InputName;
begin
  case Index of
    HrastSVF_Input: Result:='Input';
    HrastSVF_Cutoff: Result:='Cutoff';
    HrastSVF_Reso: Result:='Reso';
    HrastSVF_CutoffInput: Result:='CutoffInput';
    HrastSVF_CutoffMod: Result:='CutoffMod';
    HrastSVF_ResoInput: Result:='ResoInput';
    HrastSVF_ResoMod: Result:='ResoMod';
    HrastSVF_Gain: Result:='Gain';
    HrastSVF_Drive: Result:='Drive';
    HrastSVF_LowPass: Result:='LowPass';
    HrastSVF_BandPass: Result:='BandPass';
    HrastSVF_HighPass: Result:='HighPass';
    HrastSVF_Level12: Result:='Level12';
    HrastSVF_Level24: Result:='Level24'
    else Result:=No_Input
  end
end;

function THrastSVF.OutputName;
begin
  case Index of
    HrastSVF_OutputLP: Result:='OutputLP';
    HrastSVF_OutputBP: Result:='OutputBP';
    HrastSVF_OutputHP: Result:='OutputHP';
    HrastSVF_OutputBR: Result:='OutputBR';
    HrastSVF_Output12: Result:='Output12';
    HrastSVF_Output24: Result:='Output24'
    else Result:=No_Output
  end
end;

procedure THrastSVF.SetDefaults;
begin
  FInCache[HrastSVF_Input]:=0;
  FInCache[HrastSVF_CutoffInput]:=0;
  FInCache[HrastSVF_ResoInput]:=0;
  FKnobMap:=[
    HrastSVF_Cutoff,
    HrastSVF_Reso,
    HrastSVF_CutoffMod,
    HrastSVF_ResoMod,
    HrastSVF_Gain,
    HrastSVF_Drive,
    HrastSVF_LowPass,
    HrastSVF_BandPass,
    HrastSVF_HighPass,
    HrastSVF_Level12,
    HrastSVF_Level24
  ]
end;

procedure THrastSVF.DoTick;
var
  Oversample: Integer;
  f,fb,y,High12,High24,Output12,Output24: TSignal;
begin
  f:=Pow2Freq(FInCache[HrastSVF_Cutoff]+FInCache[HrastSVF_CutoffMod]*FInCache[HrastSVF_CutoffInput])/System_Rate_2;
  f:=2*Sine(f);
  fb:=1-(FInCache[HrastSVF_Reso]+FInCache[HrastSVF_ResoMod]*FInCache[HrastSVF_ResoInput]);
  if fb<0 then fb:=0 else if fb>1 then fb:=1;
  y:=FInCache[HrastSVF_Input]*FInCache[HrastSVF_Gain];
  y:=y/(Abs(y)+(1-FInCache[HrastSVF_Drive]));
  for Oversample:=1 to 2 do
  begin
    High12:=y-fb*Band12-Low12;
    Band12:=f*High12+Band12;
    Low12:=f*Band12+Low12;
    Low12:=Low12*1.0001/(0.0001*Abs(Low12)+1);
    Output12:=Low12*FInCache[HrastSVF_LowPass]+Band12*FInCache[HrastSVF_BandPass]+High12*FInCache[HrastSVF_HighPass];
    High24:=Output12-fb*Band24-Low24;
    Band24:=f*High24+Band24;
    Low24:=f*Band24+Low24;
    Low24:=Low24*1.0001/(0.01*Abs(Low24)+1)
  end;
  Output24:=Low24*FInCache[HrastSVF_LowPass]+Band24*FInCache[HrastSVF_BandPass]+High24*FInCache[HrastSVF_HighPass];
  FOutCache[HrastSVF_OutputLP]:=Low12;
  FOutCache[HrastSVF_OutputBP]:=Band12;
  FOutCache[HrastSVF_OutputHP]:=High12;
  FOutCache[HrastSVF_OutputBR]:=Low12+High12;
  FOutCache[HrastSVF_Output12]:=Output12*FInCache[HrastSVF_Level12];
  FOutCache[HrastSVF_Output24]:=Output24*FInCache[HrastSVF_Level24]
end;

procedure CreateTables;
var
  i: Integer;
  x: Double;
begin
  for i:=0 to TableSize do Pow2Table[i]:=UnitsToFrequency(i/TableSize);
  Pow2Table[TableSize+1]:=Pow2Table[TableSize];
  x:=Pi/TableSize; for i:=0 to TableSize-1 do SineTable[i]:=Sin(x*i);
  SineTable[TableSize div 2]:=0; SineTable[TableSize]:=SineTable[0]; SineTable[TableSize+1]:=SineTable[1];
end;

initialization

  CreateTables

end.
